#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from .config import Config
from .cursor import Cursor
from .legacy_config_transformer import LegacyConfigTransformer
from .source import SourceS3
from .stream_reader import SourceS3StreamReader

__all__ = ["Config", "Cursor", "LegacyConfigTransformer", "SourceS3", "SourceS3StreamReader"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import List, Optional, Union

from airbyte_cdk import AirbyteTracedException, FailureType

from .source_files_abstract.file_info import FileInfo


class S3Exception(AirbyteTracedException):
    def __init__(
        self,
        file_info: Union[List[FileInfo], FileInfo],
        internal_message: Optional[str] = None,
        message: Optional[str] = None,
        failure_type: FailureType = FailureType.system_error,
        exception: BaseException = None,
    ):
        file_info = (
            file_info
            if isinstance(file_info, (list, tuple))
            else [
                file_info,
            ]
        )
        file_names = ", ".join([file.key for file in file_info])
        user_friendly_message = f"""
        The connector encountered an error while processing the file(s): {file_names}.
        {message}
        This can be an input configuration error as well, please double check your connection settings.
        """
        super().__init__(internal_message=internal_message, message=user_friendly_message, failure_type=failure_type, exception=exception)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from datetime import timedelta
from typing import Any, Iterator, Mapping

import pendulum
from boto3 import session as boto3session
from botocore import UNSIGNED
from botocore.config import Config
from botocore.exceptions import ClientError

from airbyte_cdk import AirbyteTracedException, FailureType
from source_s3.s3_utils import make_s3_client

from .s3file import S3File
from .source_files_abstract.file_info import FileInfo
from .source_files_abstract.stream import IncrementalFileStream


class IncrementalFileStreamS3(IncrementalFileStream):
    @property
    def storagefile_class(self) -> type:
        return S3File

    def filepath_iterator(self, stream_state: Mapping[str, Any] = None) -> Iterator[FileInfo]:
        """
        :yield: url filepath to use in S3File()
        """
        stream_state = self._get_converted_stream_state(stream_state)
        prefix = self._provider.get("path_prefix")
        if prefix is None:
            prefix = ""

        msg = f"Iterating S3 bucket '{self._provider['bucket']}'"
        self.logger.info(msg + f" with prefix: '{prefix}' " if prefix != "" else msg)

        provider = self._provider
        client_config = None
        if S3File.use_aws_account(provider):
            session = boto3session.Session(
                aws_access_key_id=provider["aws_access_key_id"], aws_secret_access_key=provider["aws_secret_access_key"]
            )
        else:
            session = boto3session.Session()
            client_config = Config(signature_version=UNSIGNED)
        client = make_s3_client(provider, config=client_config, session=session)

        ctoken = None
        while True:
            # list_objects_v2 doesn't like a None value for ContinuationToken
            # so we don't set it if we don't have one.
            if ctoken:
                kwargs = dict(Bucket=provider["bucket"], Prefix=provider.get("path_prefix", ""), ContinuationToken=ctoken)  # type: ignore[unreachable]
            else:
                kwargs = dict(Bucket=provider["bucket"], Prefix=provider.get("path_prefix", ""))
            try:
                response = client.list_objects_v2(**kwargs)
                content = response["Contents"]
            except ClientError as e:
                message = e.response.get("Error", {}).get("Message", {})
                raise AirbyteTracedException(message, message, failure_type=FailureType.config_error)
            except KeyError:
                pass
            else:
                for file in content:
                    if self.is_not_folder(file) and self._filter_by_last_modified_date(file, stream_state):
                        yield FileInfo(key=file["Key"], last_modified=file["LastModified"], size=file["Size"])
            ctoken = response.get("NextContinuationToken", None)
            if not ctoken:
                break

    @staticmethod
    def is_not_folder(file) -> bool:
        return not file["Key"].endswith("/")

    def _filter_by_last_modified_date(self, file: Mapping[str, Any] = None, stream_state: Mapping[str, Any] = None):
        cursor_date = pendulum.parse(stream_state.get(self.cursor_field)) if stream_state else self.start_date

        file_in_history_and_last_modified_is_earlier_than_cursor_value = (
            stream_state is not None
            and self.cursor_field in stream_state.keys()
            and file.get("LastModified") <= self._get_datetime_from_stream_state(stream_state)
            and self.file_in_history(file["Key"], stream_state.get("history", {}))
        )

        file_is_not_in_history_and_last_modified_plus_buffer_days_is_earlier_than_cursor_value = file.get("LastModified") + timedelta(
            days=self.buffer_days
        ) < self._get_datetime_from_stream_state(stream_state) and not self.file_in_history(file["Key"], stream_state.get("history", {}))

        return (
            file.get("LastModified") > cursor_date
            and not file_in_history_and_last_modified_is_earlier_than_cursor_value
            and not file_is_not_in_history_and_last_modified_plus_buffer_days_is_earlier_than_cursor_value
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from __future__ import annotations

import sys
import traceback
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Mapping, Optional

import orjson

from airbyte_cdk import (
    AirbyteEntrypoint,
    ConnectorSpecification,
    emit_configuration_as_airbyte_control_message,
    is_cloud_environment,
    launch,
)
from airbyte_cdk.models import (
    AirbyteErrorTraceMessage,
    AirbyteMessage,
    AirbyteTraceMessage,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteCatalogSerializer,
    TraceType,
    Type,
)
from airbyte_cdk.sources.file_based.file_based_source import DEFAULT_CONCURRENCY, FileBasedSource
from source_s3.source import SourceS3Spec
from source_s3.utils import airbyte_message_to_json
from source_s3.v4.config import Config
from source_s3.v4.cursor import Cursor
from source_s3.v4.legacy_config_transformer import LegacyConfigTransformer
from source_s3.v4.stream_reader import SourceS3StreamReader


_V3_DEPRECATION_FIELD_MAPPING = {
    "dataset": "streams.name",
    "format": "streams.format",
    "path_pattern": "streams.globs",
    "provider": "bucket, aws_access_key_id, aws_secret_access_key and endpoint",
    "schema": "streams.input_schema",
}


class SourceS3(FileBasedSource):
    _concurrency_level = DEFAULT_CONCURRENCY

    @classmethod
    def read_config(cls, config_path: str) -> Mapping[str, Any]:
        """
        Used to override the default read_config so that when the new file-based S3 connector processes a config
        in the legacy format, it can be transformed into the new config. This happens in entrypoint before we
        validate the config against the new spec.
        """
        config = super().read_config(config_path)
        if not SourceS3._is_v4_config(config):
            parsed_legacy_config = SourceS3Spec(**config)
            converted_config = LegacyConfigTransformer.convert(parsed_legacy_config)
            emit_configuration_as_airbyte_control_message(converted_config)
            return converted_config
        return config

    def spec(self, *args: Any, **kwargs: Any) -> ConnectorSpecification:
        s3_spec = SourceS3Spec.schema()
        s4_spec = self.spec_class.schema()

        if s3_spec["properties"].keys() & s4_spec["properties"].keys():
            raise ValueError("Overlapping properties between V3 and V4")  # pragma: no cover

        for v3_property_key, v3_property_value in s3_spec["properties"].items():
            s4_spec["properties"][v3_property_key] = v3_property_value
            s4_spec["properties"][v3_property_key]["airbyte_hidden"] = True
            s4_spec["properties"][v3_property_key]["order"] += 100
            s4_spec["properties"][v3_property_key]["description"] = (
                SourceS3._create_description_with_deprecation_prefix(_V3_DEPRECATION_FIELD_MAPPING.get(v3_property_key, None))
                + s4_spec["properties"][v3_property_key]["description"]
            )
            self._clean_required_fields(s4_spec["properties"][v3_property_key])

        if is_cloud_environment():
            s4_spec["properties"]["endpoint"].update(
                {
                    "description": "Endpoint to an S3 compatible service. Leave empty to use AWS. "
                    "The custom endpoint must be secure, but the 'https' prefix is not required.",
                    "pattern": "^(?!http://).*$",  # ignore-https-check
                }
            )

        return ConnectorSpecification(
            documentationUrl=self.spec_class.documentation_url(),
            connectionSpecification=s4_spec,
        )

    @staticmethod
    def _is_v4_config(config: Mapping[str, Any]) -> bool:
        return "streams" in config

    @staticmethod
    def _clean_required_fields(v3_field: Dict[str, Any]) -> None:
        """
        Not having V3 fields root level as part of the `required` field is not enough as the platform will create empty objects for those.
        For example, filling all non-hidden fields from the form will create a config like:
        ```
        {
          <...>
          "provider": {},
          <...>
        }
        ```

        As the field `provider` exists, the JSON validation will be applied and as `provider.bucket` is needed, the validation will fail
        with the following error:
        ```
          "errors": {
            "connectionConfiguration": {
              "provider": {
                "bucket": {
                  "message": "form.empty.error",
                  "type": "required"
                }
              }
            }
          }
        ```

        Hence, we need to make any V3 nested fields not required.
        """
        if "properties" not in v3_field:
            return

        v3_field["required"] = []
        for neste_field in v3_field["properties"]:
            SourceS3._clean_required_fields(neste_field)

    @staticmethod
    def _create_description_with_deprecation_prefix(new_fields: Optional[str]) -> str:  # pragma: no cover
        if new_fields:
            return f"Deprecated and will be removed soon. Please do not use this field anymore and use {new_fields} instead. "

        return "Deprecated and will be removed soon. Please do not use this field anymore. "

    @classmethod
    def launch(cls, args: list[str] | None = None) -> None:
        """Launch the source using the provided CLI args.

        If no args are provided, the launch args will be inferred automatically.

        In the future, we should consider moving this method to the Connector base class,
        so that all sources and destinations can launch themselves and so none of this
        code needs to live in the connector itself.
        """
        args = args or sys.argv[1:]
        catalog_path = AirbyteEntrypoint.extract_catalog(args)
        # TODO: Delete if not needed:
        # config_path = AirbyteEntrypoint.extract_config(args)
        # state_path = AirbyteEntrypoint.extract_state(args)

        source = cls.create(
            configured_catalog_path=Path(catalog_path) if catalog_path else None,
        )
        # The following function will wrap the execution in proper error handling.
        # Failures prior to here may not emit proper Airbyte TRACE or CONNECTION_STATUS messages.
        launch(
            source=source,
            args=args,
        )

    @classmethod
    def create(
        cls,
        *,
        configured_catalog_path: Path | str | None = None,
    ) -> SourceS3:
        """Create a new instance of the source.

        This is a bit of a hack because (1) the source needs the catalog early, and (2), the
        constructor asks for things that the caller won't know about, specifically: the stream
        reader class, the spec class, and the cursor class.

        We should consider refactoring the constructor so that these inputs don't need to be
        provided by the caller. This probably requires changes to the base class in the CDK.

        We prefer to fail in the `launch` method, where proper error handling is in place.
        """
        try:
            configured_catalog: ConfiguredAirbyteCatalog | None = (
                ConfiguredAirbyteCatalogSerializer.load(orjson.loads(Path(configured_catalog_path).read_text()))
                if configured_catalog_path
                else None
            )
        except Exception as ex:
            print(
                airbyte_message_to_json(
                    AirbyteMessage(
                        type=Type.TRACE,
                        trace=AirbyteTraceMessage(
                            type=TraceType.ERROR,
                            emitted_at=int(datetime.now().timestamp() * 1000),
                            error=AirbyteErrorTraceMessage(
                                message="Error starting the sync. This could be due to an invalid configuration or catalog. Please contact Support for assistance.",
                                stack_trace=traceback.format_exc(),
                                internal_message=str(ex),
                            ),
                        ),
                    ),
                    newline=True,
                )
            )
            # Ideally we'd call `raise` here, but sometimes the stack trace bleeds into
            # the Airbyte logs, which is not ideal. So we'll just exit with an error code instead.
            sys.exit(1)

        return cls(
            # These are the defaults for the source. No need for a caller to change them:
            stream_reader=SourceS3StreamReader(),
            spec_class=Config,
            cursor_cls=Cursor,
            # This is needed early. (We also will provide it again later.)
            catalog=configured_catalog,
            # These will be provided later, after we have wrapped proper error handling.
            config=None,
            state=None,
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, Dict, Optional

import dpath.util
from pydantic.v1 import AnyUrl, Field, root_validator
from pydantic.v1.error_wrappers import ValidationError

from airbyte_cdk import is_cloud_environment
from airbyte_cdk.sources.file_based.config.abstract_file_based_spec import AbstractFileBasedSpec, DeliverRawFiles, DeliverRecords


class Config(AbstractFileBasedSpec):
    """
    NOTE: When this Spec is changed, legacy_config_transformer.py must also be modified to uptake the changes
    because it is responsible for converting legacy S3 v3 configs into v4 configs using the File-Based CDK.
    """

    @classmethod
    def documentation_url(cls) -> AnyUrl:
        return AnyUrl("https://docs.airbyte.com/integrations/sources/s3", scheme="https")

    bucket: str = Field(title="Bucket", description="Name of the S3 bucket where the file(s) exist.", order=0)

    aws_access_key_id: Optional[str] = Field(
        title="AWS Access Key ID",
        default=None,
        description="In order to access private Buckets stored on AWS S3, this connector requires credentials with the proper "
        "permissions. If accessing publicly available data, this field is not necessary.",
        airbyte_secret=True,
        order=2,
    )

    role_arn: Optional[str] = Field(
        title=f"AWS Role ARN",
        default=None,
        description="Specifies the Amazon Resource Name (ARN) of an IAM role that you want to use to perform operations "
        f"requested using this profile. Set the External ID to the Airbyte workspace ID, which can be found in the URL of this page.",
        order=6,
    )

    aws_secret_access_key: Optional[str] = Field(
        title="AWS Secret Access Key",
        default=None,
        description="In order to access private Buckets stored on AWS S3, this connector requires credentials with the proper "
        "permissions. If accessing publicly available data, this field is not necessary.",
        airbyte_secret=True,
        order=3,
    )

    endpoint: Optional[str] = Field(
        default="",
        title="Endpoint",
        description="Endpoint to an S3 compatible service. Leave empty to use AWS.",
        examples=["my-s3-endpoint.com", "https://my-s3-endpoint.com"],
        order=4,
    )

    region_name: Optional[str] = Field(
        title="AWS Region",
        default=None,
        description="AWS region where the S3 bucket is located. If not provided, the region will be determined automatically.",
        order=5,
    )

    delivery_method: DeliverRecords | DeliverRawFiles = Field(
        title="Delivery Method",
        discriminator="delivery_type",
        type="object",
        order=6,
        display_type="radio",
        group="advanced",
        default="use_records_transfer",
    )

    @root_validator
    def validate_optional_args(cls, values):
        aws_access_key_id = values.get("aws_access_key_id")
        aws_secret_access_key = values.get("aws_secret_access_key")
        if (aws_access_key_id or aws_secret_access_key) and not (aws_access_key_id and aws_secret_access_key):
            raise ValidationError(
                "`aws_access_key_id` and `aws_secret_access_key` are both required to authenticate with AWS.", model=Config
            )

        if is_cloud_environment():
            endpoint = values.get("endpoint")
            if endpoint:
                if endpoint.startswith("http://"):  # ignore-https-check
                    raise ValidationError("The endpoint must be a secure HTTPS endpoint.", model=Config)

        return values

    @classmethod
    def schema(cls, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """
        Generates the mapping comprised of the config fields
        """
        schema = super().schema(*args, **kwargs)

        # Hide API processing option until https://github.com/airbytehq/airbyte-platform-internal/issues/10354 is fixed
        processing_options = dpath.util.get(schema, "properties/streams/items/properties/format/oneOf/4/properties/processing/oneOf")
        dpath.util.set(schema, "properties/streams/items/properties/format/oneOf/4/properties/processing/oneOf", processing_options[:1])

        return schema



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

import io
import struct
import zipfile
from typing import IO, List, Optional, Tuple, Union

from botocore.client import BaseClient

from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from source_s3.v4.config import Config


# Buffer constants
BUFFER_SIZE_DEFAULT = 1024 * 1024
MAX_BUFFER_SIZE_DEFAULT: int = 16 * BUFFER_SIZE_DEFAULT


class RemoteFileInsideArchive(RemoteFile):
    """
    A file inside archive in a file-based stream.
    """

    start_offset: int
    compressed_size: int
    uncompressed_size: int
    compression_method: int


class ZipFileHandler:
    """
    Handler class for extracting information from ZIP files stored in AWS S3.
    """

    # Class constants for ZIP file signatures
    EOCD_SIGNATURE: bytes = b"\x50\x4b\x05\x06"
    ZIP64_LOCATOR_SIGNATURE: bytes = b"\x50\x4b\x06\x07"

    # Standard ZIP constants
    EOCD_CENTRAL_DIR_START_OFFSET: int = 16

    # ZIP64 constants
    ZIP64_EOCD_OFFSET: int = 8
    ZIP64_EOCD_SIZE: int = 56
    ZIP64_CENTRAL_DIR_START_OFFSET: int = 48

    def __init__(self, s3_client: BaseClient, config: Config):
        """
        Initialize the ZipFileHandler with an S3 client and configuration.

        :param s3_client: The AWS S3 client.
        :param config: Configuration containing bucket and other details.
        """
        self.s3_client = s3_client
        self.config = config

    def _fetch_data_from_s3(self, filename: str, start: int, size: Optional[int] = None) -> bytes:
        """
        Fetch a specific range of bytes from a file in S3.

        :param filename: The name of the file in S3.
        :param start: The starting byte position.
        :param size: The number of bytes to fetch (optional).
        :return: The fetched bytes.
        """
        end_range = f"{start + size - 1}" if size else ""
        range_str = f"bytes={start}-{end_range}"
        response = self.s3_client.get_object(Bucket=self.config.bucket, Key=filename, Range=range_str)
        return response["Body"].read()

    def _find_signature(
        self,
        filename: str,
        signature: bytes,
        initial_buffer_size: int = BUFFER_SIZE_DEFAULT,
        max_buffer_size: int = MAX_BUFFER_SIZE_DEFAULT,
    ) -> Optional[bytes]:
        """
        Search for a specific signature in the file by checking chunks of increasing size.
        If the signature is not found within the max_buffer_size, None is returned.

        :param filename: The name of the file in S3.
        :param signature: The byte signature to search for.
        :param initial_buffer_size: Initial size of the buffer to search in.
        :param max_buffer_size: Maximum size of the buffer to search in.
        :return: The chunk of data containing the signature or None if not found.
        """
        buffer_size = initial_buffer_size
        file_size = self.s3_client.head_object(Bucket=self.config.bucket, Key=filename)["ContentLength"]

        while buffer_size <= max_buffer_size:
            chunk = self._fetch_data_from_s3(filename, file_size - buffer_size)
            index = chunk.rfind(signature)
            if index != -1:
                return chunk[index:]
            buffer_size *= 2
        return None

    def _fetch_zip64_data(self, filename: str) -> bytes:
        """
        Fetch the ZIP64 End of Central Directory (EOCD) data from a ZIP file.

        :param filename: The name of the file in S3.
        :return: The ZIP64 EOCD data.
        """
        chunk = self._find_signature(filename, self.ZIP64_LOCATOR_SIGNATURE)
        zip64_eocd_offset = struct.unpack_from("<Q", chunk, self.ZIP64_EOCD_OFFSET)[0]
        return self._fetch_data_from_s3(filename, zip64_eocd_offset, self.ZIP64_EOCD_SIZE)

    def _get_central_directory_start(self, filename: str) -> int:
        """
        Determine the starting position of the central directory in the ZIP file.
        Adjusts for ZIP64 format if necessary.

        :param filename: The name of the file in S3.
        :return: The starting position of the central directory.
        """
        eocd_data = self._find_signature(filename, self.EOCD_SIGNATURE)
        central_dir_start = struct.unpack_from("<L", eocd_data, self.EOCD_CENTRAL_DIR_START_OFFSET)[0]

        # Check for ZIP64 format and adjust offsets if necessary
        if central_dir_start == 0xFFFFFFFF:
            zip64_data = self._fetch_zip64_data(filename)
            central_dir_start = struct.unpack_from("<Q", zip64_data, self.ZIP64_CENTRAL_DIR_START_OFFSET)[0]

        return central_dir_start

    def get_zip_files(self, filename: str) -> Tuple[List[zipfile.ZipInfo], int]:
        """
        Extract metadata about the files inside a ZIP archive stored in S3.

        :param filename: The name of the ZIP file in S3.
        :return: A tuple containing a list of ZipInfo objects representing the files inside the ZIP archive
                 and the starting position of the central directory.
        """
        central_dir_start = self._get_central_directory_start(filename)
        central_dir_data = self._fetch_data_from_s3(filename, central_dir_start)

        with io.BytesIO(central_dir_data) as bytes_io:
            with zipfile.ZipFile(bytes_io, "r") as zf:
                return zf.infolist(), central_dir_start


class DecompressedStream(io.IOBase):
    """
    A custom stream class that handles decompression of data from a given file object.
    This class supports seeking, reading, and other basic file operations on compressed data.
    """

    LOCAL_FILE_HEADER_SIZE: int = 30
    NAME_LENGTH_OFFSET: int = 26

    def __init__(self, file_obj: IO[bytes], file_info: RemoteFileInsideArchive, buffer_size: int = BUFFER_SIZE_DEFAULT):
        """
        Initialize a DecompressedStream.

        :param file_obj: Underlying file-like object.
        :param file_info: Meta information about the file inside the archive.
        :param buffer_size: Size of the buffer for reading data.
        """
        self._file = file_obj
        self.file_start = self._calculate_actual_start(file_info.start_offset)
        self.compressed_size = file_info.compressed_size
        self.uncompressed_size = file_info.uncompressed_size
        self.compression_method = file_info.compression_method
        self._buffer = bytearray()
        self.buffer_size = buffer_size
        self._reset_decompressor()
        self.position = 0  # Current position in uncompressed stream
        self._file.seek(self.file_start)
        # Mapping between uncompressed and compressed offsets for quick seeking
        self.offset_map = {0: self.file_start, self.uncompressed_size: self.file_start + self.compressed_size}

    def _calculate_actual_start(self, file_start: int) -> int:
        """
        Determine the actual start position of the file content within the ZIP archive.

        In a ZIP archive, each file entry is preceded by a local file header. This header contains
        metadata about the file, including the lengths of the file's name and any extra data.
        To accurately locate the start of the actual file content, we need to skip over this header.

        This method calculates the start position by taking into account the length of the file name
        and any extra data present in the local file header.

        :param file_start: The starting position of the file entry (including its local file header)
                           inside the ZIP archive.
        :return: The actual starting position of the file content, after skipping the local file header.
        """
        self._file.seek(file_start + self.NAME_LENGTH_OFFSET)  # Navigate to the position where lengths of name and extra data are stored
        name_len, extra_len = struct.unpack("<HH", self._file.read(4))  # Extract the lengths
        return file_start + self.LOCAL_FILE_HEADER_SIZE + name_len + extra_len  # Calculate the actual start by skipping the header

    def _reset_decompressor(self):
        """
        Reset the decompressor object.
        """
        self.decompressor = zipfile._get_decompressor(self.compression_method)

    def _decompress_chunk(self, chunk: bytes) -> bytes:
        """
        Decompress a chunk of data based on the compression method.
        """
        if self.compression_method == zipfile.ZIP_STORED:
            return chunk
        return self.decompressor.decompress(chunk)

    def read(self, size: int = -1) -> bytes:
        """
        Read a specified number of bytes from the stream.
        """
        # Size not specified, read till end
        if size == -1:
            size = self.uncompressed_size - self.position

        # If buffer already has enough data, return it directly
        if size <= len(self._buffer):
            data = self._buffer[:size]
            self._buffer = self._buffer[size:]
            self.position += len(data)
            return data

        data = self._buffer
        self._buffer = bytearray()
        while len(data) < size and self._file.tell() - self.file_start < self.compressed_size:
            max_read_size = min(self.buffer_size, self.compressed_size + self.file_start - self._file.tell())
            chunk = self._file.read(max_read_size)

            if not chunk:
                break

            decompressed_data = self._decompress_chunk(chunk)

            # Buffer excessive data for future reads
            if len(data) + len(decompressed_data) > size:
                desired_length = size - len(data)
                data += decompressed_data[:desired_length]
                self._buffer = decompressed_data[desired_length:]
            else:
                data += decompressed_data

        self.position += len(data)
        return data

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """
        Seek to a specific position in the uncompressed stream.
        """
        if whence == io.SEEK_SET:
            self._buffer = bytearray()
        elif whence == io.SEEK_CUR:
            offset = self.position + offset
        elif whence == io.SEEK_END:
            offset = self.uncompressed_size + offset

        # Ensure the offset is within the file's boundaries
        offset = max(0, min(offset, self.uncompressed_size))

        closest_offset = max(k for k in self.offset_map if k <= offset)
        closest_position = self.offset_map[closest_offset]

        self._file.seek(closest_position)
        self._reset_decompressor()
        self.position = closest_offset

        # Read till desired offset
        while self.position < offset:
            read_size = min(self.buffer_size, offset - self.position)
            self.read(read_size)

        return self.position

    def tell(self) -> int:
        """
        Return the current position in the uncompressed stream.
        """
        return self.position

    def readable(self) -> bool:
        """
        Return if the stream is readable.
        """
        return True

    def seekable(self) -> bool:
        """
        Return if the stream is seekable.
        """
        return True

    def close(self):
        """
        Close the stream and underlying file object.
        """
        self._file.close()


class ZipContentReader:
    """
    A custom reader class that provides buffered reading capabilities on a decompressed stream.
    Supports reading lines, reading chunks, and iterating over the content.
    """

    def __init__(self, decompressed_stream: DecompressedStream, encoding: Optional[str] = None, buffer_size: int = BUFFER_SIZE_DEFAULT):
        """
        Initialize a ZipContentReader.

        :param decompressed_stream: A DecompressedStream object.
        :param encoding: Encoding to decode the bytes. If None, bytes are returned.
        :param buffer_size: Size of the buffer for reading data.
        """
        self.raw = decompressed_stream
        self.encoding = encoding
        self.buffer_size = buffer_size
        self.buffer = bytearray()
        self._closed = False

    def __iter__(self):
        """
        Make the class iterable.
        """
        return self

    def __next__(self) -> Union[str, bytes]:
        """
        Iterate over the lines in the reader.
        """
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    def readline(self, limit: int = -1) -> Union[str, bytes]:
        """
        Read a single line from the stream.
        """
        if limit != -1:
            raise NotImplementedError("Limits other than -1 not implemented yet")

        line = ""
        while True:
            char = self.read(1)
            if not char:
                break

            line += char
            if char in ["\n", "\r"]:
                # Handling different types of newlines
                next_char = self.read(1)
                if char == "\r" and next_char == "\n":
                    line += next_char
                else:
                    self.buffer = next_char.encode(self.encoding) + self.buffer
                break
        return line

    def read(self, size: int = -1) -> Union[str, bytes]:
        """
        Read a specified number of bytes/characters from the reader.
        """
        while len(self.buffer) < size:
            chunk = self.raw.read(self.buffer_size)
            if not chunk:
                break
            self.buffer += chunk

        data = self.buffer[:size]
        self.buffer = self.buffer[size:]

        return data.decode(self.encoding) if self.encoding else bytes(data)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """
        Seek to a specific position in the decompressed stream.
        """
        self.buffer = bytearray()
        return self.raw.seek(offset, whence)

    def close(self):
        """
        Close the reader and underlying decompressed stream.
        """
        self._closed = True
        self.raw.close()

    def tell(self) -> int:
        """
        Return the current position in the decompressed stream.
        """
        return self.raw.tell()

    @property
    def closed(self) -> bool:
        """
        Check if the reader is closed.
        """
        return self._closed

    def __enter__(self) -> "ZipContentReader":
        """Enter the runtime context for the reader."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the runtime context for the reader and ensure resources are closed."""
        self.close()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import json
import re
from typing import Any, Dict, Union

from jsonschema import RefResolver
from pydantic.v1 import BaseModel, Field

from .formats.avro_spec import AvroFormat
from .formats.csv_spec import CsvFormat
from .formats.jsonl_spec import JsonlFormat
from .formats.parquet_spec import ParquetFormat


# To implement your provider specific spec, inherit from SourceFilesAbstractSpec and add provider-specific settings e.g.:

# class SourceS3Spec(SourceFilesAbstractSpec, BaseModel):
#     class Config:
#         title="S3 Source Spec"

#     class S3Provider(BaseModel):
#         class Config:
#             title = "S3: Amazon Web Services"

#         bucket: str = Field(description="Name of the S3 bucket where the file(s) exist.")
#         aws_access_key_id: Optional[str] = Field(
#             default=None,
#             description="...",
#             airbyte_secret=True
#         )
#         aws_secret_access_key: Optional[str] = Field(
#             default=None,
#             description="...",
#             airbyte_secret=True
#         )
#         path_prefix: str = Field(
#             default="",
#             description="..."
#         )
#     provider: S3Provider = Field(...)  # leave this as Field(...), just change type to relevant class


class SourceFilesAbstractSpec(BaseModel):
    dataset: str = Field(
        pattern=r"^([A-Za-z0-9-_]+)$",
        description="The name of the stream you would like this source to output. Can contain letters, numbers, or underscores.",
        order=0,
        title="Output Stream Name",
    )

    path_pattern: str = Field(
        title="Pattern of files to replicate",
        description="A regular expression which tells the connector which files to replicate. All files which match this pattern will be "
        'replicated. Use | to separate multiple patterns. See <a href="https://facelessuser.github.io/wcmatch/glob/" target="_'
        'blank">this page</a> to understand pattern syntax (GLOBSTAR and SPLIT flags are enabled). '
        "Use pattern <strong>**</strong> to pick up all files.",
        examples=["**", "myFolder/myTableFiles/*.csv|myFolder/myOtherTableFiles/*.csv"],
        order=10,
    )

    format: Union[CsvFormat, ParquetFormat, AvroFormat, JsonlFormat] = Field(
        default="csv", title="File Format", description="The format of the files you'd like to replicate", order=20
    )

    user_schema: str = Field(
        title="Manually enforced data schema",
        alias="schema",
        default="{}",
        description="Optionally provide a schema to enforce, as a valid JSON string. Ensure this is a mapping of "
        '<strong>{ "column" : "type" }</strong>, where types are valid '
        '<a href="https://json-schema.org/understanding-json-schema/reference/type.html" target="_blank">JSON Schema '
        "datatypes</a>. Leave as {} to auto-infer the schema.",
        examples=['{"column_1": "number", "column_2": "string", "column_3": "array", "column_4": "object", "column_5": "boolean"}'],
        order=30,
    )

    @staticmethod
    def change_format_to_oneOf(schema: dict) -> dict:
        props_to_change = ["format"]
        for prop in props_to_change:
            schema["properties"][prop]["type"] = "object"
            if "oneOf" in schema["properties"][prop]:
                continue
            schema["properties"][prop]["oneOf"] = schema["properties"][prop].pop("anyOf")
        return schema

    @staticmethod
    def remove_enum_allOf(schema: dict) -> dict:
        """
        allOfs are not supported by the UI, but pydantic is automatically writing them for enums.
        Unpack them into the root
        """
        objects_to_check = schema["properties"]["format"]["oneOf"]
        for object in objects_to_check:
            for key in object["properties"]:
                property = object["properties"][key]
                if "allOf" in property and "enum" in property["allOf"][0]:
                    property["enum"] = property["allOf"][0]["enum"]
                    property.pop("allOf")
        return schema

    @staticmethod
    def check_provider_added(schema: dict) -> None:
        if "provider" not in schema["properties"]:
            raise RuntimeError("You must add the 'provider' property in your child spec class")

    @staticmethod
    def resolve_refs(schema: dict) -> dict:
        json_schema_ref_resolver = RefResolver.from_schema(schema)
        str_schema = json.dumps(schema)
        for ref_block in re.findall(r'{"\$ref": "#\/definitions\/.+?(?="})"}', str_schema):
            ref = json.loads(ref_block)["$ref"]
            str_schema = str_schema.replace(ref_block, json.dumps(json_schema_ref_resolver.resolve(ref)[1]))
        pyschema: dict = json.loads(str_schema)
        del pyschema["definitions"]
        return pyschema

    @classmethod
    def schema(cls, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """we're overriding the schema classmethod to enable some post-processing"""
        schema = super().schema(*args, **kwargs)
        cls.check_provider_added(schema)
        schema = cls.change_format_to_oneOf(schema)
        schema = cls.resolve_refs(schema)
        schema = cls.remove_enum_allOf(schema)
        return schema



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
import time
from datetime import datetime
from io import IOBase
from os import getenv, makedirs, path
from typing import Dict, Iterable, List, Optional, Set, cast

import boto3.session
import pendulum
import psutil
import pytz
import smart_open
from botocore.client import BaseClient
from botocore.client import Config as ClientConfig
from botocore.credentials import RefreshableCredentials
from botocore.exceptions import ClientError
from botocore.session import get_session
from typing_extensions import override

from airbyte_cdk import FailureType
from airbyte_cdk.sources.file_based.exceptions import CustomFileBasedException, ErrorListingFiles, FileBasedSourceError, FileSizeLimitError
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader, FileReadMode
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from source_s3.v4.config import Config
from source_s3.v4.zip_reader import DecompressedStream, RemoteFileInsideArchive, ZipContentReader, ZipFileHandler


AWS_EXTERNAL_ID = getenv("AWS_ASSUME_ROLE_EXTERNAL_ID")


class SourceS3StreamReader(AbstractFileBasedStreamReader):
    FILE_SIZE_LIMIT = 1_500_000_000

    def __init__(self):
        super().__init__()
        self._s3_client = None

    @property
    def config(self) -> Config:
        return self._config

    @config.setter
    def config(self, value: Config):
        """
        FileBasedSource reads the config from disk and parses it, and once parsed, the source sets the config on its StreamReader.

        Note: FileBasedSource only requires the keys defined in the abstract config, whereas concrete implementations of StreamReader
        will require keys that (for example) allow it to authenticate with the 3rd party.

        Therefore, concrete implementations of AbstractFileBasedStreamReader's config setter should assert that `value` is of the correct
        config type for that type of StreamReader.
        """
        assert isinstance(value, Config)
        self._config = value

    @property
    def s3_client(self) -> BaseClient:
        if self.config is None:
            # We shouldn't hit this; config should always get set before attempting to
            # list or read files.
            raise ValueError("Source config is missing; cannot create the S3 client.")

        if self._s3_client is None:
            client_kv_args = _get_s3_compatible_client_args(self.config) if self.config.endpoint else {}

            # Set the region_name if it's provided in the config
            if self.config.region_name:
                client_kv_args["region_name"] = self.config.region_name

            if self.config.role_arn:
                self._s3_client = self._get_iam_s3_client(client_kv_args)
            else:
                self._s3_client = boto3.client(
                    "s3",
                    aws_access_key_id=self.config.aws_access_key_id,
                    aws_secret_access_key=self.config.aws_secret_access_key,
                    **client_kv_args,
                )

        return self._s3_client

    def _get_iam_s3_client(self, client_kv_args: dict) -> BaseClient:
        """
        Creates an S3 client using AWS Security Token Service (STS) with assumed role credentials. This method handles
        the authentication process by assuming an IAM role, optionally using an external ID for enhanced security.
        The obtained credentials are set to auto-refresh upon expiration, ensuring uninterrupted access to the S3 service.

        :param client_kv_args: A dictionary of key-value pairs for the boto3 S3 client constructor.
        :return: An instance of a boto3 S3 client with the assumed role credentials.

        The method assumes a role specified in the `self.config.role_arn` and creates a session with the S3 service.
        If `AWS_ASSUME_ROLE_EXTERNAL_ID` environment variable is set, it will be used during the role assumption for additional security.
        """

        def refresh():
            client = boto3.client("sts")
            if AWS_EXTERNAL_ID:
                role = client.assume_role(
                    RoleArn=self.config.role_arn,
                    RoleSessionName="airbyte-source-s3",
                    ExternalId=AWS_EXTERNAL_ID,
                )
            else:
                role = client.assume_role(
                    RoleArn=self.config.role_arn,
                    RoleSessionName="airbyte-source-s3",
                )

            creds = role.get("Credentials", {})
            return {
                "access_key": creds["AccessKeyId"],
                "secret_key": creds["SecretAccessKey"],
                "token": creds["SessionToken"],
                "expiry_time": creds["Expiration"].isoformat(),
            }

        session_credentials = RefreshableCredentials.create_from_metadata(
            metadata=refresh(),
            refresh_using=refresh,
            method="sts-assume-role",
        )

        session = get_session()
        session._credentials = session_credentials
        autorefresh_session = boto3.Session(botocore_session=session)

        return autorefresh_session.client("s3", **client_kv_args)

    def get_matching_files(self, globs: List[str], prefix: Optional[str], logger: logging.Logger) -> Iterable[RemoteFile]:
        """
        Get all files matching the specified glob patterns.
        """
        s3 = self.s3_client
        prefixes = [prefix] if prefix else self.get_prefixes_from_globs(globs)
        seen = set()
        total_n_keys = 0

        try:
            for current_prefix in prefixes if prefixes else [None]:
                for remote_file in self._page(s3, globs, self.config.bucket, current_prefix, seen, logger):
                    total_n_keys += 1
                    yield remote_file

            logger.info(f"Finished listing objects from S3. Found {total_n_keys} objects total ({len(seen)} unique objects).")
        except ClientError as exc:
            if exc.response["Error"]["Code"] == "NoSuchBucket":
                raise CustomFileBasedException(
                    f"The bucket {self.config.bucket} does not exist.", failure_type=FailureType.config_error, exception=exc
                )
            self._raise_error_listing_files(globs, exc)
        except Exception as exc:
            self._raise_error_listing_files(globs, exc)

    def _raise_error_listing_files(self, globs: List[str], exc: Optional[Exception] = None):
        """Helper method to raise the ErrorListingFiles exception."""
        raise ErrorListingFiles(
            FileBasedSourceError.ERROR_LISTING_FILES,
            source="s3",
            bucket=self.config.bucket,
            globs=globs,
            endpoint=self.config.endpoint,
        ) from exc

    def open_file(self, file: RemoteFile, mode: FileReadMode, encoding: Optional[str], logger: logging.Logger) -> IOBase:
        try:
            params = {"client": self.s3_client}
        except Exception as exc:
            raise exc

        logger.debug(f"try to open {file.uri}")
        try:
            if isinstance(file, RemoteFileInsideArchive):
                s3_file_object = smart_open.open(f"s3://{self.config.bucket}/{file.uri.split('#')[0]}", transport_params=params, mode="rb")
                decompressed_stream = DecompressedStream(s3_file_object, file)
                result = ZipContentReader(decompressed_stream, encoding)
            else:
                result = smart_open.open(
                    f"s3://{self.config.bucket}/{file.uri}", transport_params=params, mode=mode.value, encoding=encoding
                )
        except OSError:
            logger.warning(
                f"We don't have access to {file.uri}. The file appears to have become unreachable during sync."
                f"Check whether key {file.uri} exists in `{self.config.bucket}` bucket and/or has proper ACL permissions"
            )

        # we can simply return the result here as it is a context manager itself that will release all resources
        return result

    @staticmethod
    def create_progress_handler(file_size: int, local_file_path: str, logger: logging.Logger):
        previous_bytes_checkpoint = 0
        total_bytes_transferred = 0

        def progress_handler(bytes_transferred: int):
            nonlocal previous_bytes_checkpoint, total_bytes_transferred
            total_bytes_transferred += bytes_transferred
            if total_bytes_transferred - previous_bytes_checkpoint >= 100 * 1024 * 1024:
                logger.info(
                    f"{total_bytes_transferred / (1024 * 1024):,.2f} MB ({total_bytes_transferred / (1024 * 1024 * 1024):.2f} GB) "
                    f"of {file_size / (1024 * 1024):,.2f} MB ({file_size / (1024 * 1024 * 1024):.2f} GB) "
                    f"written to {local_file_path}"
                )
                previous_bytes_checkpoint = total_bytes_transferred

                # Get available disk space
                disk_usage = psutil.disk_usage("/")
                available_disk_space = disk_usage.free

                # Get available memory
                memory_info = psutil.virtual_memory()
                available_memory = memory_info.available
                logger.info(
                    f"Available disk space: {available_disk_space / (1024 * 1024):,.2f} MB ({available_disk_space / (1024 * 1024 * 1024):.2f} GB), "
                    f"available memory: {available_memory / (1024 * 1024):,.2f} MB ({available_memory / (1024 * 1024 * 1024):.2f} GB)."
                )

        return progress_handler

    @override
    def get_file(self, file: RemoteFile, local_directory: str, logger: logging.Logger) -> Dict[str, str | int]:
        """
        Downloads a file from an S3 bucket to a specified local directory.

        Args:
            file (RemoteFile): The remote file object containing URI and metadata.
            local_directory (str): The local directory path where the file will be downloaded.
            logger (logging.Logger): Logger for logging information and errors.

        Returns:
            dict: A dictionary containing the following:
                - "file_url" (str): The absolute path of the downloaded file.
                - "bytes" (int): The file size in bytes.
                - "file_relative_path" (str): The relative path of the file for local storage. Is relative to local_directory as
                this a mounted volume in the pod container.

        Raises:
            FileSizeLimitError: If the file size exceeds the predefined limit (1 GB).
        """
        file_size = self.file_size(file)
        # I'm putting this check here so we can remove the safety wheels per connector when ready.
        if file_size > self.FILE_SIZE_LIMIT:
            message = "File size exceeds the 1 GB limit."
            raise FileSizeLimitError(message=message, internal_message=message, failure_type=FailureType.config_error)

        file_relative_path, local_file_path, absolute_file_path = self._get_file_transfer_paths(file, local_directory)

        logger.info(
            f"Starting to download the file {file.uri} with size: {file_size / (1024 * 1024):,.2f} MB ({file_size / (1024 * 1024 * 1024):.2f} GB)"
        )
        # at some moment maybe we will require to play with the max_pool_connections and max_concurrency of s3 config
        start_download_time = time.time()
        progress_handler = self.create_progress_handler(file_size, local_file_path, logger)
        self.s3_client.download_file(self.config.bucket, file.uri, local_file_path, Callback=progress_handler)
        write_duration = time.time() - start_download_time
        logger.info(f"Finished downloading the file {file.uri} and saved to {local_file_path} in {write_duration:,.2f} seconds.")

        return {"file_url": absolute_file_path, "bytes": file_size, "file_relative_path": file_relative_path}

    @override
    def file_size(self, file: RemoteFile) -> int:
        s3_object: boto3.s3.Object = self.s3_client.get_object(
            Bucket=self.config.bucket,
            Key=file.uri,
        )
        return cast(int, s3_object["ContentLength"])

    @staticmethod
    def _is_folder(file) -> bool:
        return file["Key"].endswith("/")

    def _page(
        self, s3: BaseClient, globs: List[str], bucket: str, prefix: Optional[str], seen: Set[str], logger: logging.Logger
    ) -> Iterable[RemoteFile]:
        """
        Page through lists of S3 objects.
        """
        total_n_keys_for_prefix = 0
        kwargs = {"Bucket": bucket}
        while True:
            response = s3.list_objects_v2(Prefix=prefix, **kwargs) if prefix else s3.list_objects_v2(**kwargs)
            key_count = response.get("KeyCount")
            total_n_keys_for_prefix += key_count
            logger.info(f"Received {key_count} objects from S3 for prefix '{prefix}'.")

            if "Contents" in response:
                for file in response["Contents"]:
                    if self._is_folder(file):
                        continue

                    for remote_file in self._handle_file(file):
                        if (
                            self.file_matches_globs(remote_file, globs)
                            and self.is_modified_after_start_date(remote_file.last_modified)
                            and remote_file.uri not in seen
                        ):
                            seen.add(remote_file.uri)
                            yield remote_file
            else:
                logger.warning(f"Invalid response from S3; missing 'Contents' key. kwargs={kwargs}.")

            if next_token := response.get("NextContinuationToken"):
                kwargs["ContinuationToken"] = next_token
            else:
                logger.info(f"Finished listing objects from S3 for prefix={prefix}. Found {total_n_keys_for_prefix} objects.")
                break

    def is_modified_after_start_date(self, last_modified_date: Optional[datetime]) -> bool:
        """Returns True if given date higher or equal than start date or something is missing"""
        if not (self.config.start_date and last_modified_date):
            return True
        return last_modified_date >= pendulum.parse(self.config.start_date).naive()

    def _handle_file(self, file):
        if file["Key"].endswith(".zip"):
            yield from self._handle_zip_file(file)
        else:
            yield self._handle_regular_file(file)

    def _handle_zip_file(self, file):
        zip_handler = ZipFileHandler(self.s3_client, self.config)
        zip_members, cd_start = zip_handler.get_zip_files(file["Key"])

        for zip_member in zip_members:
            remote_file = RemoteFileInsideArchive(
                uri=file["Key"] + "#" + zip_member.filename,
                last_modified=datetime(*zip_member.date_time).astimezone(pytz.utc).replace(tzinfo=None),
                start_offset=zip_member.header_offset + cd_start,
                compressed_size=zip_member.compress_size,
                uncompressed_size=zip_member.file_size,
                compression_method=zip_member.compress_type,
            )
            yield remote_file

    def _handle_regular_file(self, file):
        remote_file = RemoteFile(uri=file["Key"], last_modified=file["LastModified"].astimezone(pytz.utc).replace(tzinfo=None))
        return remote_file


def _get_s3_compatible_client_args(config: Config) -> dict:
    """
    Returns map of args used for creating s3 boto3 client.
    """
    client_kv_args = {
        "config": ClientConfig(s3={"addressing_style": "auto"}),
        "endpoint_url": config.endpoint,
        "use_ssl": True,
        "verify": True,
    }
    return client_kv_args



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from enum import Enum

from pydantic.v1 import BaseModel, Field


class UnexpectedFieldBehaviorEnum(str, Enum):
    ignore = "ignore"
    infer = "infer"
    error = "error"


class JsonlFormat(BaseModel):
    'This connector uses <a href="https://arrow.apache.org/docs/python/json.html" target="_blank">PyArrow</a> for JSON Lines (jsonl) file parsing.'

    class Config:
        title = "Jsonl"

    filetype: str = Field(
        "jsonl",
        const=True,
    )

    newlines_in_values: bool = Field(
        title="Allow newlines in values",
        default=False,
        description="Whether newline characters are allowed in JSON values. Turning this on may affect performance. Leave blank to default to False.",
        order=0,
    )

    unexpected_field_behavior: UnexpectedFieldBehaviorEnum = Field(
        title="Unexpected field behavior",
        default="infer",
        description='How JSON fields outside of explicit_schema (if given) are treated. Check <a href="https://arrow.apache.org/docs/python/generated/pyarrow.json.ParseOptions.html" target="_blank">PyArrow documentation</a> for details',
        examples=["ignore", "infer", "error"],
        order=1,
    )
    # Block size set to 0 as default value to disable this feature for most not-experienced users
    block_size: int = Field(
        default=0,
        description="The chunk size in bytes to process at a time in memory from each file. If your data is particularly wide and failing during schema detection, increasing this should solve it. Beware of raising this too high as you could hit OOM errors.",
        order=2,
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
from datetime import datetime
from typing import Any, Dict, List, Mapping, Optional, Union

from source_s3.source import SourceS3Spec
from source_s3.source_files_abstract.formats.avro_spec import AvroFormat
from source_s3.source_files_abstract.formats.csv_spec import CsvFormat
from source_s3.source_files_abstract.formats.jsonl_spec import JsonlFormat
from source_s3.source_files_abstract.formats.parquet_spec import ParquetFormat


SECONDS_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
MICROS_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class LegacyConfigTransformer:
    """
    Class that takes in S3 source configs in the legacy format and transforms them into
    configs that can be used by the new S3 source built with the file-based CDK.
    """

    @classmethod
    def convert(cls, legacy_config: SourceS3Spec) -> Mapping[str, Any]:
        transformed_config = {
            "bucket": legacy_config.provider.bucket,
            "streams": [
                {
                    "name": legacy_config.dataset,
                    "globs": cls._create_globs(legacy_config.path_pattern),
                    "legacy_prefix": legacy_config.provider.path_prefix,
                    "validation_policy": "Emit Record",
                }
            ],
        }

        if legacy_config.provider.start_date:
            transformed_config["start_date"] = cls._transform_seconds_to_micros(legacy_config.provider.start_date)
        if legacy_config.provider.aws_access_key_id:
            transformed_config["aws_access_key_id"] = legacy_config.provider.aws_access_key_id
        if legacy_config.provider.aws_secret_access_key:
            transformed_config["aws_secret_access_key"] = legacy_config.provider.aws_secret_access_key
        if legacy_config.provider.endpoint:
            transformed_config["endpoint"] = legacy_config.provider.endpoint
        if legacy_config.user_schema and legacy_config.user_schema != "{}":
            transformed_config["streams"][0]["input_schema"] = legacy_config.user_schema
        if legacy_config.format:
            transformed_config["streams"][0]["format"] = cls._transform_file_format(legacy_config.format)

        return transformed_config

    @classmethod
    def _create_globs(cls, path_pattern: str) -> List[str]:
        if "|" in path_pattern:
            return path_pattern.split("|")
        else:
            return [path_pattern]

    @classmethod
    def _transform_seconds_to_micros(cls, datetime_str: str) -> str:
        try:
            parsed_datetime = datetime.strptime(datetime_str, SECONDS_FORMAT)
            return parsed_datetime.strftime(MICROS_FORMAT)
        except ValueError as e:
            raise ValueError("Timestamp could not be parsed when transforming legacy connector config") from e

    @classmethod
    def _transform_file_format(cls, format_options: Union[CsvFormat, ParquetFormat, AvroFormat, JsonlFormat]) -> Mapping[str, Any]:
        if isinstance(format_options, AvroFormat):
            return {"filetype": "avro"}
        elif isinstance(format_options, CsvFormat):
            additional_reader_options = cls.parse_config_options_str("additional_reader_options", format_options.additional_reader_options)
            advanced_options = cls.parse_config_options_str("advanced_options", format_options.advanced_options)

            csv_options = {
                "filetype": "csv",
                "delimiter": format_options.delimiter,
                "quote_char": format_options.quote_char,
                "double_quote": format_options.double_quote,
                # values taken from https://github.com/apache/arrow/blob/43c05c56b37daa93e76b94bc3e6952d56d1ea3f2/cpp/src/arrow/csv/options.cc#L41-L45
                "null_values": additional_reader_options.pop(
                    "null_values",
                    [
                        "",
                        "#N/A",
                        "#N/A N/A",
                        "#NA",
                        "-1.#IND",
                        "-1.#QNAN",
                        "-NaN",
                        "-nan",
                        "1.#IND",
                        "1.#QNAN",
                        "N/A",
                        "NA",
                        "NULL",
                        "NaN",
                        "n/a",
                        "nan",
                        "null",
                    ],
                ),
                "true_values": additional_reader_options.pop("true_values", ["1", "True", "TRUE", "true"]),
                "false_values": additional_reader_options.pop("false_values", ["0", "False", "FALSE", "false"]),
                "inference_type": "Primitive Types Only" if format_options.infer_datatypes else "None",
                "strings_can_be_null": additional_reader_options.pop("strings_can_be_null", False),
            }

            if format_options.escape_char:
                csv_options["escape_char"] = format_options.escape_char
            if format_options.encoding:
                csv_options["encoding"] = format_options.encoding
            if skip_rows := advanced_options.pop("skip_rows", None):
                csv_options["skip_rows_before_header"] = skip_rows
            if skip_rows_after_names := advanced_options.pop("skip_rows_after_names", None):
                csv_options["skip_rows_after_header"] = skip_rows_after_names

            if column_names := advanced_options.pop("column_names", None):
                csv_options["header_definition"] = {
                    "header_definition_type": "User Provided",
                    "column_names": column_names,
                }
                advanced_options.pop("autogenerate_column_names", None)
            elif advanced_options.pop("autogenerate_column_names", None):
                csv_options["header_definition"] = {"header_definition_type": "Autogenerated"}
            else:
                csv_options["header_definition"] = {"header_definition_type": "From CSV"}

            cls._filter_legacy_noops(advanced_options)

            if advanced_options or additional_reader_options:
                raise ValueError(
                    "The config options you selected are no longer supported.\n" + f"advanced_options={advanced_options}"
                    if advanced_options
                    else "" + f"additional_reader_options={additional_reader_options}"
                    if additional_reader_options
                    else ""
                )

            return csv_options

        elif isinstance(format_options, JsonlFormat):
            return {"filetype": "jsonl"}
        elif isinstance(format_options, ParquetFormat):
            return {"filetype": "parquet", "decimal_as_float": True}
        else:
            # This should never happen because it would fail schema validation
            raise ValueError(f"Format filetype {format_options} is not a supported file type")

    @classmethod
    def parse_config_options_str(cls, options_field: str, options_value: Optional[str]) -> Dict[str, Any]:
        options_str = options_value or "{}"
        try:
            return json.loads(options_str)
        except json.JSONDecodeError as error:
            raise ValueError(f"Malformed {options_field} config json: {error}. Please ensure that it is a valid JSON.")

    @staticmethod
    def _filter_legacy_noops(advanced_options: Dict[str, Any]):
        ignore_all = ("auto_dict_encode", "timestamp_parsers", "block_size")
        ignore_by_value = (("check_utf8", False),)

        for option in ignore_all:
            advanced_options.pop(option, None)

        for option, value_to_ignore in ignore_by_value:
            if advanced_options.get(option) == value_to_ignore:
                advanced_options.pop(option)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Optional

from pydantic.v1 import BaseModel, Field

from .source_files_abstract.spec import SourceFilesAbstractSpec


class SourceS3Spec(SourceFilesAbstractSpec, BaseModel):
    class Config:
        title = "S3 Source Spec"

    class S3Provider(BaseModel):
        class Config:
            title = "S3: Amazon Web Services"
            # SourceFilesAbstractSpec field are ordered 10 apart to allow subclasses to insert their own spec's fields interspersed
            schema_extra = {"order": 11, "description": "Use this to load files from S3 or S3-compatible services"}

        bucket: str = Field(description="Name of the S3 bucket where the file(s) exist.", order=0)
        aws_access_key_id: Optional[str] = Field(
            title="AWS Access Key ID",
            default=None,
            description="In order to access private Buckets stored on AWS S3, this connector requires credentials with the proper "
            "permissions. If accessing publicly available data, this field is not necessary.",
            airbyte_secret=True,
            always_show=True,
            order=1,
        )
        aws_secret_access_key: Optional[str] = Field(
            title="AWS Secret Access Key",
            default=None,
            description="In order to access private Buckets stored on AWS S3, this connector requires credentials with the proper "
            "permissions. If accessing publicly available data, this field is not necessary.",
            airbyte_secret=True,
            always_show=True,
            order=2,
        )
        role_arn: Optional[str] = Field(
            title=f"AWS Role ARN",
            default=None,
            description="Specifies the Amazon Resource Name (ARN) of an IAM role that you want to use to perform operations "
            f"requested using this profile. Set the External ID to the Airbyte workspace ID, which can be found in the URL of this page.",
            always_show=True,
            order=7,
        )
        path_prefix: str = Field(
            default="",
            description="By providing a path-like prefix (e.g. myFolder/thisTable/) under which all the relevant files sit, "
            "we can optimize finding these in S3. This is optional but recommended if your bucket contains many "
            "folders/files which you don't need to replicate.",
            order=3,
        )

        endpoint: str = Field("", description="Endpoint to an S3 compatible service. Leave empty to use AWS.", order=4)
        region_name: Optional[str] = Field(
            title="AWS Region",
            default=None,
            description="AWS region where the S3 bucket is located. If not provided, the region will be determined automatically.",
            order=5,
        )
        start_date: Optional[str] = Field(
            title="Start Date",
            description="UTC date and time in the format 2017-01-25T00:00:00Z. Any file modified before this date will not be replicated.",
            examples=["2021-01-01T00:00:00Z"],
            format="date-time",
            pattern="^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$",
            order=6,
        )

    provider: S3Provider



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from datetime import datetime, timedelta
from typing import Any, MutableMapping

from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.stream.cursor import DefaultFileBasedCursor
from airbyte_cdk.sources.file_based.types import StreamState


logger = logging.Logger("source-S3")


class Cursor(DefaultFileBasedCursor):
    _DATE_FORMAT = "%Y-%m-%d"
    _LEGACY_DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
    _V4_MIGRATION_BUFFER = timedelta(hours=1)
    _V3_MIN_SYNC_DATE_FIELD = "v3_min_sync_date"

    def __init__(self, stream_config: FileBasedStreamConfig, **_: Any):
        super().__init__(stream_config)
        self._running_migration = False
        self._v3_migration_start_datetime = None

    def set_initial_state(self, value: StreamState) -> None:
        if self._is_legacy_state(value):
            self._running_migration = True
            value = self._convert_legacy_state(value)
        else:
            self._running_migration = False
        self._v3_migration_start_datetime = (
            datetime.strptime(value.get(Cursor._V3_MIN_SYNC_DATE_FIELD), DefaultFileBasedCursor.DATE_TIME_FORMAT)
            if Cursor._V3_MIN_SYNC_DATE_FIELD in value
            else None
        )
        super().set_initial_state(value)

    def get_state(self) -> StreamState:
        state = {"history": self._file_to_datetime_history, self.CURSOR_FIELD: self._get_cursor()}
        if self._v3_migration_start_datetime:
            return {
                **state,
                **{
                    Cursor._V3_MIN_SYNC_DATE_FIELD: datetime.strftime(
                        self._v3_migration_start_datetime, DefaultFileBasedCursor.DATE_TIME_FORMAT
                    )
                },
            }
        else:
            return state

    def _should_sync_file(self, file: RemoteFile, logger: logging.Logger) -> bool:
        """
        Never sync files earlier than the v3 migration start date. V3 purged the history from the state, so we assume all files were already synced
        Else if the currenty sync is migrating from v3 to v4, sync all files that were modified within one hour of the last sync
        Else sync according to the default logic
        """
        if self._v3_migration_start_datetime and file.last_modified < self._v3_migration_start_datetime:
            return False
        elif self._running_migration:
            return True
        else:
            return super()._should_sync_file(file, logger)

    @staticmethod
    def _is_legacy_state(value: StreamState) -> bool:
        if not value:
            return False
        try:
            # Verify datetime format in history
            history = value.get("history", {}).keys()
            if history:
                item = list(value.get("history", {}).keys())[0]
                datetime.strptime(item, Cursor._DATE_FORMAT)

            # verify the format of the last_modified cursor
            last_modified_at_cursor = value.get(DefaultFileBasedCursor.CURSOR_FIELD)
            if not last_modified_at_cursor:
                return False
            datetime.strptime(last_modified_at_cursor, Cursor._LEGACY_DATE_TIME_FORMAT)
        except ValueError:
            return False
        return True

    @staticmethod
    def _convert_legacy_state(legacy_state: StreamState) -> MutableMapping[str, Any]:
        """
        Transform the history from the old state message format to the new.

        e.g.
        {
            "2022-05-26": ["simple_test.csv.csv", "simple_test_2.csv"],
            "2022-05-27": ["simple_test_2.csv", "redshift_result.csv"],
            ...
        }
        =>
        {
            "simple_test.csv": "2022-05-26T00:00:00.000000Z",
            "simple_test_2.csv": "2022-05-27T00:00:00.000000Z",
            "redshift_result.csv": "2022-05-27T00:00:00.000000Z",
            ...
        }
        """
        converted_history = {}
        legacy_cursor = legacy_state[DefaultFileBasedCursor.CURSOR_FIELD]
        cursor_datetime = datetime.strptime(legacy_cursor, Cursor._LEGACY_DATE_TIME_FORMAT)
        logger.info(f"Converting v3 -> v4 state. v3_cursor={legacy_cursor} v3_history={legacy_state.get('history')}")

        for date_str, filenames in legacy_state.get("history", {}).items():
            datetime_obj = Cursor._get_adjusted_date_timestamp(cursor_datetime, datetime.strptime(date_str, Cursor._DATE_FORMAT))

            for filename in filenames:
                if filename in converted_history:
                    if datetime_obj > datetime.strptime(
                        converted_history[filename],
                        DefaultFileBasedCursor.DATE_TIME_FORMAT,
                    ):
                        converted_history[filename] = datetime_obj.strftime(DefaultFileBasedCursor.DATE_TIME_FORMAT)
                    else:
                        # If the file was already synced with a later timestamp, ignore
                        pass
                else:
                    converted_history[filename] = datetime_obj.strftime(DefaultFileBasedCursor.DATE_TIME_FORMAT)

        if converted_history:
            filename, _ = max(converted_history.items(), key=lambda x: (x[1], x[0]))
            cursor = f"{cursor_datetime}_{filename}"
        else:
            # Having a cursor with empty history is not expected, but we handle it.
            logger.warning(f"Cursor found without a history object; this is not expected. cursor_value={legacy_cursor}")
            # Note: we convert to the v4 cursor granularity, but since no items are in the history we simply use the
            # timestamp as the cursor value instead of the concatenation of timestamp_filename, which is the v4
            # cursor format.
            # This is okay because the v4 cursor is kept for posterity but is not actually used in the v4 code. If we
            # start to use the cursor we may need to revisit this logic.
            cursor = cursor_datetime
            converted_history = {}
        v3_migration_start_datetime = cursor_datetime - Cursor._V4_MIGRATION_BUFFER
        return {
            "history": converted_history,
            DefaultFileBasedCursor.CURSOR_FIELD: cursor,
            Cursor._V3_MIN_SYNC_DATE_FIELD: v3_migration_start_datetime.strftime(DefaultFileBasedCursor.DATE_TIME_FORMAT),
        }

    @staticmethod
    def _get_adjusted_date_timestamp(cursor_datetime: datetime, file_datetime: datetime) -> datetime:
        if file_datetime > cursor_datetime:
            return file_datetime
        else:
            # Extract the dates so they can be compared
            cursor_date = cursor_datetime.date()
            date_obj = file_datetime.date()

            # If same day, update the time to the cursor time
            if date_obj == cursor_date:
                return file_datetime.replace(hour=cursor_datetime.hour, minute=cursor_datetime.minute, second=cursor_datetime.second)
            # If previous, update the time to end of day
            else:
                return file_datetime.replace(hour=23, minute=59, second=59, microsecond=999999)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from __future__ import annotations

import logging
import multiprocessing as mp
import traceback
from multiprocessing import Queue
from typing import Any, Callable, List, Mapping, cast

import dill
import orjson

from airbyte_cdk.models import AirbyteMessage, AirbyteMessageSerializer


def run_in_external_process(fn: Callable, timeout: int, max_timeout: int, logger: logging.Logger, args: List[Any]) -> Mapping[str, Any]:
    """
    fn passed in must return a tuple of (desired return value, Exception OR None)
    This allows propagating any errors from the process up and raising accordingly
    """
    result = None
    while result is None:
        q_worker: Queue = mp.Queue()
        proc = mp.Process(
            target=multiprocess_queuer,
            # use dill to pickle the function for Windows-compatibility
            args=(dill.dumps(fn), q_worker, *args),
        )
        proc.start()
        try:
            # this attempts to get return value from function with our specified timeout up to max
            result, potential_error = q_worker.get(timeout=min(timeout, max_timeout))
        except mp.queues.Empty:  # type: ignore[attr-defined]
            if timeout >= max_timeout:  # if we've got to max_timeout and tried once with that value
                raise TimeoutError(f"Timed out too many times while running {fn.__name__}, max timeout of {max_timeout} seconds reached.")
            logger.info(f"timed out while running {fn.__name__} after {timeout} seconds, retrying...")
            timeout *= 2  # double timeout and try again
        else:
            if potential_error is None:
                return result  # type: ignore[no-any-return]
            traceback.print_exception(type(potential_error), potential_error, potential_error.__traceback__)
            raise potential_error
        finally:
            try:
                proc.terminate()
            except Exception as e:
                logger.info(f"'{fn.__name__}' proc unterminated, error: {e}")


def multiprocess_queuer(func: Callable, queue: mp.Queue, *args: Any, **kwargs: Any) -> None:
    """this is our multiprocesser helper function, lives at top-level to be Windows-compatible"""
    queue.put(dill.loads(func)(*args, **kwargs))


def get_value_or_json_if_empty_string(options: str = None) -> str:
    return options.strip() if options else "{}"


def airbyte_message_to_json(
    message: AirbyteMessage,
    *,
    newline: bool = False,
) -> str:
    """Dump the provided AirbyteMessage to a JSON string.

    Optionally append a newline character to the end of the string.
    """
    result = orjson.dumps(cast(dict, AirbyteMessageSerializer.dump(message))).decode()
    if newline:
        result += "\n"

    return result


def airbyte_message_from_json(message_json: str) -> AirbyteMessage:
    """Create an AirbyteMessage object from the provided JSON string."""
    return AirbyteMessageSerializer.load(orjson.loads(message_json))
